


package go;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import go.*;
import antlr.TokenStreamException;
import antlr.RecognitionException;

import e.*;
import libGo.*;

class EventQE
{
    int          type;
    EventQE      next;    
    int          x, y;
    char         key;
    int          funcIndex;
    int          argCount;
    double[]     arg;
    String       script;
    String       scriptName;

    EventQE()
    {
	next = null;
    }
}

public abstract class GoInterfaceE extends Canvas implements Runnable
{
    final static int RENDER          = 1;
    final static int SIZE            = 2;
    final static int KEY_TYPED       = 3;
    final static int MOUSE_PRESSED   = 4;
    final static int MOUSE_RELEASED  = 5;
    final static int MOUSE_DRAGGED   = 6;
    final static int COMPILE_EXECUTE = 7;
    final static int EXECUTE_CODE    = 8;
    final static int EXECUTE_FUNC    = 9;

    // Thread
    public Thread thread = null;

    // Event Queue
    public EventQE       head = new EventQE();
    EventQE              tail = head;

    // E variables
    public E            e;
    public boolean      scriptGood  = false;
    public ExecuteFunc  executeFunc = new ExecuteFunc();
    PopStatic           popStatic = new PopStatic();

    // Event function index values
    public int  renderIndex;
    public int  sizeIndex;
    public int  keyTypedIndex;
    public int  mousePressedIndex;
    public int  mouseReleasedIndex;
    public int  mouseDraggedIndex;

    /** Variable containing the Go context that was created with this interface.
     */
    public Go  go;

    //
    // Java image variables.
    //
    MemoryImageSource source;
    Image             image = null;
    int               width = 1;
    int               height = 1;
    boolean           dirty = false;

    // Arguments passed to e's event functions.
    public double[] sizeArg = new double[2];
    public double[] keyTypedArg = new double[1];
    public double[] mousePressedArg = new double[2];
    public double[] mouseReleasedArg = new double[2];
    public double[] mouseDraggedArg = new double[2];

    public GoInterfaceE()
    {
        //
        // Create a new Go context.
        //
        initializeGo();

	//
	// Resized event listener.
	//
        addComponentListener
	    (
	     new ComponentAdapter()
		 {
		     public void componentResized(ComponentEvent event)
		     {
			 width  = getSize().width;
			 height = getSize().height;

			 // clamp
			 if(width < 1) width = 1;
			 if(height < 1) height = 1;
			 
			 tail = tail.next = new EventQE();
			 tail.type = SIZE;
			 tail.x = width;
			 tail.y = height;

			 notifyRenderThread();
		     }
		 }
	     );

	//
	// Keyboard event listener.
	//
	addKeyListener
            (
	     new KeyAdapter()
		 {
		     public synchronized void keyTyped(KeyEvent event)
		     {
			 tail = tail.next = new EventQE();
			 tail.type = KEY_TYPED;
			 tail.key = event.getKeyChar();

			 notifyRenderThread();
		     }
		 }
	     );

	//
	// Mouse pressed and released event listener.
	//
	addMouseListener
            (
	     new MouseAdapter()
		 {
		     public synchronized void mousePressed(MouseEvent event)
		     {
			 tail = tail.next = new EventQE();
			 tail.type = MOUSE_PRESSED;
			 tail.x = event.getX();
			 tail.y = event.getY();

			 notifyRenderThread();
		     }
                
		     public synchronized void mouseReleased(MouseEvent event)
		     {
			 tail = tail.next = new EventQE();
			 tail.type = MOUSE_RELEASED;
			 tail.x = event.getX();
			 tail.y = event.getY();

			 notifyRenderThread();
		     }
		 }
	     );
        
	//
	// Mouse dragged event listener.
	//
	addMouseMotionListener
            (
	     new MouseMotionAdapter()
		 {
		     public synchronized void mouseDragged(MouseEvent event)
		     {
			 tail = tail.next = new EventQE();
			 tail.type = MOUSE_DRAGGED;
			 tail.x = event.getX();
			 tail.y = event.getY();

			 notifyRenderThread();
		     }
		 }
	     );
    }

    public synchronized void initEventFuncIndex()
    {
	// Initialize event function indexes to invalid.

	renderIndex = -1;
	sizeIndex = -1;
	keyTypedIndex = -1;
	mousePressedIndex = -1;
	mouseReleasedIndex = -1;
	mouseDraggedIndex = -1;
    }

    public synchronized void initEventQE()
    {
	head = new EventQE();
	tail = head;
	
	initEventFuncIndex();
    }

    public void initializeGo()
    {
	//
        // Greate a new Go context.
        //
        go = new Go(width, height);

        //
        // Create the Java image.
        //
        source = new MemoryImageSource(width, height,
                                       go.image.buf, 0, width);
        source.setAnimated(true);
        image = createImage(source);
    }
    
    public void rerender()
    {
        dirty = true;
        repaint();
    }

    public void render() throws ExecuteException
    {
	if(renderIndex >= 0) {
	    addExecuteFunc(renderIndex, 0, null);
	}
    }

    public void size(int width, int height) throws ExecuteException
    {
	e.width = width;
	e.height = height;

	if(sizeIndex >= 0) {
	    sizeArg[0] = width;
	    sizeArg[1] = height;
	    addExecuteFunc(sizeIndex, 2, sizeArg);
	}
    }

    public void keyTyped(char key) throws ExecuteException
    {
	if(keyTypedIndex >= 0) {
	    keyTypedArg[0] = key;
	    addExecuteFunc(keyTypedIndex, 1, keyTypedArg);
	}
    }

    public void mousePressed(int x, int y) throws ExecuteException
    {
	if(mousePressedIndex >= 0) {
	    mousePressedArg[0] = x;
	    mousePressedArg[1] = y;
	    addExecuteFunc(mousePressedIndex, 2, mousePressedArg);
	}
    }

    public void mouseReleased(int x, int y) throws ExecuteException
    {
	if(mouseReleasedIndex >= 0) {
	    mouseReleasedArg[0] = x;
	    mouseReleasedArg[1] = y;
	    addExecuteFunc(mouseReleasedIndex, 2, mouseReleasedArg);
	}
    }

    public void mouseDragged(int x, int y) throws ExecuteException
    {
	if(mouseDraggedIndex >= 0) {
	    mouseDraggedArg[0] = x;
	    mouseDraggedArg[1] = y;
	    addExecuteFunc(mouseDraggedIndex, 2, mouseDraggedArg);
	}
    }

    public void swap()
    {
        Graphics g = getGraphics();

        if(g != null) {

            //
            // Update the Java image.
            //
            source.newPixels(0, 0, width, height);

            //
            // Display the Java image.
            //
            g.drawImage(image, 0, 0, null);
        }
    }

    // Called upon a successful compile.
    public void compileSuccess()
    {
	// override if desired
    }

    // Execute code.
    public void execute() throws ExecuteException
    {
	e.execute();
    }

    // Compiler message.
    public void compileMessage(String message)
    {
	e.stdout(message);
    }

    // Compiler error message.
    public void compileError(String file, int line, String message)
    {
	e.stdout("Compile error in " + file + " on line " + line + ": " + message);
    }

    // Runtime error message.
    public void runtimeError(String file, int line, String message)
    {
	e.stdout("Runtime error in " + file + " on line " + line + ": " + message);
    }

    // Process ExecuteException
    synchronized void doExecuteException(ExecuteException exception, EventQE event)
    {
	event.next = null;
	initEventFuncIndex();

	scriptGood = false;

	if(!exception.normalTermination) {
	    runtimeError(exception.file, exception.lineNo, exception.getMessage());
	}
    }

    //
    // Main render loop.
    //
    public synchronized void run()
    {
	String script = null;
	String scriptName = null;

	while(true) {
	    
	    //
	    // Go to sleep if there are no events to process.
	    //

	    try {
		while(head.next == null) {
		    wait();
		}

		//
		// Post saved COMPILE_EXECUTE event.
		//
		// See "Postpone COMPILE_EXECUTE event?" comment
		// below for more information.
		//
		if(script != null) {
		    addCompileExecute(script, scriptName);
		    script = null;
		    scriptName = null;
		}
	    }
	    catch (InterruptedException exception) { /* REVISIT: I think no action is OK */ }


	    //
	    // Not sleeping, and ready to process events.
	    //

	    EventQE event = head.next;
	    
	    while(event != null) {
	    
		switch(event.type) {
	
		case RENDER:
		    try {
			render();
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }
		    break;

		case SIZE:
		    try {

			//
			// Resize go.
			//
			go.size(width, height);
			
			//
			// Resize the Java image.
			//
			source = new MemoryImageSource(width, height,
						       go.image.buf, 0, width);
			source.setAnimated(true);
			image = createImage(source);
			
			size(event.x, event.y);
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }
		    break;

		case KEY_TYPED:
		    try {
			keyTyped(event.key);
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }
		    break;

		case MOUSE_PRESSED:
		    try {
			mousePressed(event.x, event.y);
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }
		    break;

		case MOUSE_RELEASED:
		    try {
			mouseReleased(event.x, event.y);
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }
		    break;

		case MOUSE_DRAGGED:
		    try {
			mouseDragged(event.x, event.y);
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }
		    break;

		case COMPILE_EXECUTE:

		    //
		    // Postpone COMPILE_EXECUTE event?
		    // 
		    // Make sure that the render area has been defined.
		    // If not, then save away the script & scriptName
		    // and append a COMPILE_EXECUTE event later.
		    // See the "Post saved COMPILE_EXECUTE event" comment
		    // above.
		    //
		    // NOTE: Internet Explorer renders nothing if this
                    //       is not done.
		    //
		    if(e.width == 0 || e.height == 0) {
			script = event.script;
			scriptName = event.scriptName;
			break;
		    }
		    
		    // cleanup
		    if(scriptGood) {
			popStatic.execute();
			e.popTable();
			LibFunction.init();
		    }
		    
		    scriptGood = false;
		    
		    try {
			
			initEventFuncIndex();
			
			e.compile(event.script, event.scriptName);
			
			initializeGo();
			GoLibFunc.go = go;
			
			compileSuccess();
			
			// setup width and height of display before executing anything
			e.width = getSize().width;
			e.height = getSize().height;
			
			//
			// start code execution
			//
			//    - global variables will be declared and initialized here
			//
			addExecuteCode();
			
			//
			// execute 'main' function
			//
			try {
			    addExecuteFunc(e.getFunctionIndex("main"), 0, null);
			    scriptGood = true;
			}
			catch(ExecuteException exception) {
			    doExecuteException(exception, event);
			}
		    }
		    catch(TokenStreamException exception) {
			compileError(e.lexer.getFilename(), e.lexer.getLine(), exception.getMessage());
		    }
		    catch(RecognitionException exception) {
			compileError(e.lexer.getFilename(), exception.line, exception.getMessage());
		    }

		    break;

		case EXECUTE_CODE:
		    
		    try {
			execute();
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }
		    
		    break;

		case EXECUTE_FUNC:
		    
		    e.pushStack(event.funcIndex);
		    for(int i = 0; i < event.argCount; i++) {
			e.pushStack(event.arg[i]);
		    }
		    e.pushStack(event.argCount);

		    try {
			executeFunc.execute();
		    }
		    catch(ExecuteException exception) {
			doExecuteException(exception, event);
		    }

		    // After a function has been called, the stack
		    // always needs to be poped.  However, here we
                    // ignore the return value.
		    e.popStack();
		    
		    break;
		}

		event = event.next;
	    }
	    
	    head.next = null;
	    tail = head;
	}
    }

    //
    // Add compile and execute event.
    //
    public synchronized void addCompileExecute(String script, String scriptName)
    {
	tail = tail.next = new EventQE();
	tail.type = COMPILE_EXECUTE;
	tail.script = script;
	tail.scriptName = scriptName;

	notifyRenderThread();
    }

    //
    // Add initial code execution.
    //
    public synchronized void addExecuteCode()
    {
	tail = tail.next = new EventQE();
	tail.type = EXECUTE_CODE;

	notifyRenderThread();
    }

    //
    // Add function code execution.
    //
    public synchronized void addExecuteFunc(int funcIndex, int argCount, double[] arg) throws ExecuteException
    {
	if(e.func[funcIndex].argNum != argCount) {
	    if(argCount == 1) {
		// singular
		throw new ExecuteException(false, e.func[funcIndex].file, e.func[funcIndex].line,
					   "function requires " + argCount + " argument");
	    }
	    else {
		// plural
		throw new ExecuteException(false, e.func[funcIndex].file, e.func[funcIndex].line,
					   "function requires " + argCount + " arguments");
	    }
	}

	tail = tail.next = new EventQE();
	tail.type = EXECUTE_FUNC;
	tail.funcIndex = funcIndex;
	tail.argCount = argCount;
	tail.arg = arg;
	
	notifyRenderThread();
    }

    //
    // Thread notification that an event has arrived.
    //
    synchronized void notifyRenderThread()
    {
	notifyAll();
    }

    //
    // Overrides update in class java.awt.Component
    //
    public void update(Graphics g)
    {
        paint(g);
    }

    //
    // Overrides paint in class java.awt.Canvas
    //
    public void paint(Graphics g)
    {
        //
        // If dirty, then rerender with Go.
        //
        if(dirty) {
            dirty = false;
            tail = tail.next = new EventQE();
	    tail.type = RENDER;
	    notifyRenderThread();
        }

        //
        // Display the Java image.
        //
        g.drawImage(image, 0, 0, null);
    }
}
